import osmnx as ox
import networkx as nx
import folium
import pickle
import time, datetime
import os.path
import sys
import webbrowser

import os

import matplotlib.pyplot as plt
plt.switch_backend('agg')


ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
print('NetworkX version: ',nx.__version__)
print('Python version: ',sys.version)

start_time = time.time()


def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)




def report(Q, txt=''):
    print(len(Q.nodes()), ' nodes ', txt)
    print(len(Q.edges()), ' edges ', txt)
    if isinstance(Q,nx.MultiGraph):    
        length = ox.stats.edge_length_total(Q) #in meters
        length = length * 0.0006213712
        print(f'total length {length:.2f} miles ', txt)
    zeroes = 0    
    odds = 0
    evens = 0
    twos = 0
    ones = 0
    fours = 0
    four_pairs = 0
    parallels = 0
    for node in Q.nodes:
        d = Q.degree(node)
        if d == 0:
            zeroes += 1
        if d % 2 == 1:
            odds += 1
        else:
            evens += 1
        if d == 2:
            twos += 1
        if d == 1:
            ones += 1
        if d==4:
            fours += 1
            for neighbor in Q.neighbors(node):
                if Q.degree(neighbor)==4 and neighbor>node: # prevent double-count
                    four_pairs += 1
        if d>6:
            print('node ', node, 'has degree ', d)
        for neighbor in Q.neighbors(node):
            edge_count = Q.number_of_edges(node,neighbor)
            if edge_count > 1:
                parallels += edge_count - 1
                
    if zeroes > 0:            
        print('number of zero-degree nodes ' + txt + ' :', zeroes)            
    print('number of odd intersections ' + txt + ' :', odds)
    print('number of even intersections ' + txt + ' :', evens)
    print('number of 2-way intersections ' + txt + ' :', twos)
    print('number of 4-way intersections ' + txt + ' :', fours)
    print('number of trailhead/terminus ' + txt + ' :', ones)
    print('number of parallel edges ' + txt + ' :', parallels)
    if four_pairs > 0:
        print('number of 4-way adjacent pairs ' + txt + ' :', four_pairs)

        
    check_node_geometry(Q)
    print('')

def colorize_graph(Q):
    QR = Q.copy()
    for node, dat in Q.nodes(data=True):
        d = Q.degree(node)
        dat['color'] = d

        QR.add_node(node,**dat)                #replace node value

    for u, v, k, dat in Q.edges(keys=True,data=True):
        d = 'path'
        if 'highway' in dat:
            d = dat['highway']
        dat['color']=0.0
        if isinstance(d, str):
            if d=='path' or d == 'footway':
                dat['color']=1.0
            else:
                dat['color'] = 2.0
        QR.add_edge(u,v,key=k,**dat)
    QR.graph['color']=True
    return QR


def check_node_geometry(Q):
    count = 0
    for node, dat in Q.nodes(data=True):
        if 'x' not in dat:
            count += 1
            #d = Q.degree(node)
            #print('node ', node, ' has no geometry attribute, degree ', d)t
    if count>0:
        print('number of nodes without geometry: ',count)
     
town_markers = [(34.29963, -108.13268, 'Pie Town'),
         (33.19593, -108.20834, 'Gila Hot Springs'),
         (32.77981, -108.27533, 'Silver City'),
         (33.37939, -108.90369, 'Alma/Glenwood'),
         (33.03022, -108.16864, 'Lake Roberts'),
         (32.85594, -107.98024, 'Mimbres'),       
         (32.91685, -107.70593, 'Kingston')]


# draw() changed after moving to Linux
def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=town_markers):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
#    Q1,Q2 = split_merged(Q)
    T = Q.copy()
    if not 'color' in T.graph:
        T = colorize_graph(T)
    QD = T.to_directed()
    QD.graph['crs'] = ox.settings.default_crs


    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, edge_attrs_differ=['name'], endpoint_attrs=None, remove_rings=True, track_merged=True)
#        QD = ox.simplify_graph(QD, edge_attrs_differ=None, endpoint_attrs=None, remove_rings=True, track_merged=True)
#        QD = simplify_graph_modified(QD)
    QD = colorize_graph(QD)

    
    nodes, edges = ox.graph_to_gdfs(QD)
    #edges.info()
    #print(edges)
#    m = edges.explore(color='#ff00ff', tiles="openstreetmap",tooltip='name')
    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap",tooltip='name')
#    m = edges.explore(column='color',  cmap = 'tab10',legend=False, tiles="openstreetmap")
    if show_nodes:    
#        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    
        nodes.explore(m=m, column='color', cmap='tab10', legend=False,marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(n,w),(s,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          icon=folium.Icon(color='lightred',icon=''),
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
   # folium.TileLayer('openstreetmap').add_to(m)
    folium.TileLayer('cartodbpositron').add_to(m)
   # folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
    #m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    
        

def distance(Q, node, node2): #assumes we are in polar coordinates
    y1 = Q.nodes[node]['y']
    x1 = Q.nodes[node]['x']
    y2 = Q.nodes[node2]['y']
    x2 = Q.nodes[node2]['x']
    #d = ox.distance.euclidean(y1, x1, y2, x2)
    # we are in lat/long, so use different function
    d = ox.distance.great_circle(y1,x1, y2, x2) # should return distance in meters
    return d


# read graph
# Using network_type='walk' seems to include dirt roads, so use a custom filter
cf_add_routes = '["highway"~"path|footway|motorway|motorway_link|trunk|primary|secondary|tertiary|unclassified|residential|living_street|residential|service|track"]'

bbox_gila = (34.35704, 32.77935, -107.65983, -108.94574) # rough box of Gila NF

gila_trail_plus_road_filename = 'gila_trail_plus_road_from_bbox.pkl'

GTR =  ox.graph_from_bbox(bbox=bbox_gila,simplify=False,retain_all=True,custom_filter=cf_add_routes)

# add one edge to GTR that is missing from OSM as of current date
# code taken from road_and_gap_v2.py
road_list = [
    ( (33.39053, -108.78010), (33.40938, -108.77280), '+ log canyon trail'),
]
for point, point2, description in road_list:
    node = ox.nearest_nodes(GTR,point[1], point[0])
    node2 = ox.nearest_nodes(GTR,point2[1], point2[0])
    if description.startswith('+'):  #oops, missing trail, add a straight line
        GTR.add_edge(node,node2,length = distance(GTR,node,node2), name = description)


pickle.dump(GTR,open(gila_trail_plus_road_filename, 'wb'))
print('finished importing trail plus road graph')
elapsed_time(start_time)



J = pickle.load(open('max_candidate_with_towns.pkl', 'rb'))
report(J,'candidate with towns')



K = nx.MultiGraph()     
if nx.is_eulerian(J):
    print('input is eulerian!!!!!')
else:
    print('back to the drawing board')
    draw(J,'back to drawing board',show_nodes=True)


# -----------------DESIMPLIFY -------------------------------

for u,v,k,d in J.edges(keys=True,data=True):
    if 'merged_edges' in d:
        merged_edges = d['merged_edges']
    else:
        merged_edges = [ [u,v] ]

    if len(merged_edges)==0:
        continue

    for s,t in merged_edges:
        if GTR.has_edge(s,t):
            for kg in GTR[s][t]:  #fix improve
                dat = GTR[s][t][kg]
                K.add_edge(s,t,key=None,**dat)
                break
        elif GTR.has_edge(t,s): # in case 1-way divided street around Silver City
            for kg in GTR[t][s]:  #fix improve
                dat = GTR[t][s][kg]
                K.add_edge(t,s,key=None,**dat)
                break
        else:
            print('FATAL ERROR WITH MERGED EDGES',s,t) 
            continue


#copy node attributes
KR = K.copy()
for node, dat in KR.nodes(data=True):
    node_dic = GTR.nodes[node] # node really ought to be in GTR...
    K.add_node(node,** node_dic )
    

def fix_node_geometry(Q,Q_original):
    QR = Q.copy()
    for node in Q.nodes():  # add node attributes back in
        node_dic = Q.nodes[node]
        if 'x' not in node_dic:
            node_dic = Q_original.nodes[node]
            QR.add_node(node,**node_dic)
    return QR
    
K = fix_node_geometry(K,J)

def find_longest_edge(Q,u,v):
    max_k = 0
    max_length = 0
    for k in Q[u][v]:
        length = Q[u][v][k]['length']
        if length > max_length:
            max_length = length
            max_k = k
    return max_k

def remove_parallel_edges(Q):
    QR = nx.MultiGraph()
    for u,v in Q.edges():
        k = find_longest_edge(Q,u,v)
        dat = Q[u][v][k]
        QR.add_edge(u,v,k,**dat)
    QR = fix_node_geometry(QR,Q)
    return QR
    
K = remove_parallel_edges(K)

report(K,'K = J-desimplified?')              

if nx.is_eulerian(K):
    print('desimplified graph is still eulerian!!!!!')
    draw(K,'desimplified',show_nodes=False)
else:
    print('desimplified graph not eulerian-- debugging needed')

pickle.dump(K,open('max_candidate_with_towns_unsimplified.pkl', 'wb'))

K2 = pickle.load(open('max_candidate_with_towns_unsimplified.pkl', 'rb'))
if nx.is_eulerian(K2):
    print('desimplified graph loaded from disk is still eulerian')
else:
    print('desimplified graph loaded from disk not eulerian-- debugging needed')



elapsed_time(start_time)

print('finished script ' + os.path.basename(__file__))






